/**
  ##############################################################################
  @file     eeprom.c
  @brief    Работаем с eeprom
  @note     eeprom_file_init читаем из памяти файл, если его нет то пишем свой
  @note 		в eeprom_buf храним данные вычитанные 
  @note     потом в eeprom_periodic_work сравниваем eeprom_buf и регистры модбаса
  @note     записывая в eepom значение в случае расхождения
  ##############################################################################
*/


#include "eeprom.h"
#include "i2c.h"
#include "modbus_regs_file.h"
#include "state_machine.h"
#include "modbus_config.h"

#define EEPROM_BUF_SIZE 50u //в байтах
#define I2C_WRITE_BUFER_SIZE 4u
#define EEPROM_ADR 0x50

static uint8_t init_data_check(uint8_t* data);
static bool eeprom_lock_read(uint16_t addr,uint8_t* data, uint8_t length);
static uint8_t init_end_data_check(uint8_t* data);
static void eeprom_write_file(void);
static bool write_regs(uint8_t* data);
static void eeprom_write_default_file(void);

static void save_reg_state(state_machine_dev* dev);
static void comparison_regs_state(state_machine_dev* dev);

static uint8_t i2c_write_bufer[I2C_WRITE_BUFER_SIZE];
static uint8_t reg_num; //количество регистров в файле
static uint8_t eeprom_buf[EEPROM_BUF_SIZE]; //зеркало i2c файла

static state_machine_dev st_dev = {plug_state, plug_state, 0u };

/**
  @brief  отправляем новое значение регистра 
*/
static void save_reg_state(state_machine_dev* dev)
{
	if(I2C_NOT_BUSY==i2c_write(EEPROM_ADR, &i2c_write_bufer, I2C_WRITE_BUFER_SIZE))
	{
		set_state_after_delay(comparison_regs_state,2u,&st_dev);
		//xprintf("eeprom write\n");
	}
	return;
}

/**
  @brief  сравниваем eeprom_buf и регистры модбаса
*/
static void comparison_regs_state(state_machine_dev* dev)
{
	for(uint8_t i=0;i<(reg_num*ONE_REG_SIZE);i+=ONE_REG_SIZE)
	{
		uint16_t reg_addr = (eeprom_buf[i]<<8)+eeprom_buf[i+1];
		uint16_t buf_data = (eeprom_buf[i+3]<<8)+eeprom_buf[i+4];
		uint8_t reg_type = eeprom_buf[i+2];
		uint16_t reg_data;
		MBRF_read_reg(reg_addr, &reg_data, reg_type);

		//xprintf("reg write  %x %d %x\n",reg_addr, reg_type, reg_data );
		if(buf_data!=reg_data)
		{
			eeprom_buf[i+3] = (uint8_t)reg_data>>8;
			eeprom_buf[i+4] = (uint8_t)reg_data;

			i2c_write_bufer[0]=0u;
			i2c_write_bufer[1]=i+LENGTH_INIT_DATA+3;
			i2c_write_bufer[2]=eeprom_buf[i+3];
			i2c_write_bufer[3]=eeprom_buf[i+4];
			set_state(save_reg_state, &st_dev);
			save_reg_state(&st_dev); //сразу выполняем
			//xprintf("reg write %x %d %x\n",reg_addr, reg_type, reg_data);
			break;
		}
	}
}

/**
  @brief  работа
*/
void eeprom_periodic_work(void)
{
	execute_current_state(&st_dev);
	return;
}

/**
  @brief  блокирущее чтение
  @param addr адресс начала чтения 
  @param data указатель на массивкуда пишем из eeprom
  @param length кол-во регистров
  @return 0 в случае ошибки
*/
static bool eeprom_lock_read(uint16_t addr,uint8_t* data, uint8_t length)
{
	bool ret = I2C_ERROR;
	static uint8_t data_addr[2];
	data_addr[0]=(uint8_t)addr>>8;
	data_addr[1]=(uint8_t)addr;

	if(I2C_OK==i2c_lock_write(EEPROM_ADR, &data_addr, 2u))
	{
		ret = i2c_lock_read(EEPROM_ADR,data,length);
	}
	return ret;
}

/**
  @brief  проверяем наличие файла
*/
void eeprom_file_init(void)
{
	if(I2C_OK==eeprom_lock_read(0u, &eeprom_buf,LENGTH_INIT_DATA)) //если не получили NACK
	{
		reg_num = init_data_check(&eeprom_buf);
		if(0u!=reg_num)
		{
			eeprom_lock_read(LENGTH_INIT_DATA+(reg_num*ONE_REG_SIZE), &eeprom_buf,LENGTH_END_DATA);
			if(init_end_data_check(&eeprom_buf))
			{
				eeprom_lock_read((LENGTH_INIT_DATA), &eeprom_buf,reg_num*ONE_REG_SIZE);
				if(write_regs(&eeprom_buf))
				{
					#ifdef DEBUG	
					xprintf("eeprom init end\n");
    			#endif
				}
			}
		}else{
			eeprom_write_default_file(); //если нету ничего запишем стандартные настройки
			eeprom_file_init(); //плохое решение, переписать
			#ifdef DEBUG	
			xprintf("eeprom default write\n");
    	#endif
		}
		set_state(comparison_regs_state, &st_dev);
	}
	#ifdef DEBUG
	else {
		#ifdef DEBUG	
		xprintf("eeprom NACK\n");
		#endif
	}
	#endif
}

/**
  @brief записывает значения в модбас регистры
  @param уазатель на буфер с прочитанным файлом (в data[0] должны лежать данные регистра)
  @retun 0 в случае если нашли регистр который  не занесён в модбас файл
  @retun 1 в случае если всё хорошо
  @note в случае return 0 данные регистров до неизвечтного адресса будут записаны
*/
static bool write_regs(uint8_t* data)
{
	bool ret =0u;
	for(uint8_t i=0;i<(reg_num*ONE_REG_SIZE);i+=ONE_REG_SIZE)
	{
		uint16_t reg_addr = (data[i]<<8)+data[i+1];
		uint16_t reg_data = (data[i+3]<<8)+data[i+4];
		uint8_t reg_type = data[i+2];

		if(MBR_OK!=MBRF_write_reg(reg_addr, reg_data, reg_type))
		{
			#ifdef DEBUG	
			xprintf("reg write eror %x %d %x\n",reg_addr, reg_type, reg_data);
			#endif
			break;
		}
		ret =1u;
	}
	return ret;
}

/**
  @brief проверяем начало напяти на наличие ключевых битов
*/
static uint8_t init_data_check(uint8_t* data)
{
	uint8_t ret =0u;
	if((data[0]==EEPROM_S_DATA_FLAG)&&(data[1]==EEPROM_S_DATA_FLAG))
  {
    if(FIRMWARE_VERSION==data[2]) //пока это бесполезно, задел на будущее
    {
    	#ifdef DEBUG
      xprintf("FIRMWARE_VERSION 0x%02x\n",data[2]);
      #endif
      uint8_t reg_num = data[3];
      if(reg_num<(EEPROM_BUF_SIZE/ONE_REG_SIZE))
      {
      	ret=reg_num;
      	#ifdef DEBUG
      	xprintf("file_size %d\n",reg_num);
      	#endif
      }
    }
  }
	return ret;
}

/**
  @brief проверяем конец файла на наличие ключевых битов
*/
static uint8_t init_end_data_check(uint8_t* data)
{
	uint8_t ret =0u;
	if((data[0]==EEPROM_E_DATA_FLAG)&&(data[1]==EEPROM_E_DATA_FLAG))
  {
  	#ifdef DEBUG
    xprintf("init_end_data_check successfully\n");
    #endif
    ret = 1u;
  }
  return ret;
}

/**
  @brief записываем дефолтные настройки
*/
static void eeprom_write_default_file(void)
{
	uint8_t buf[] = {DEFAULT_FILE_START_ADDR0,DEFAULT_FILE_START_ADDR1,DEFAULT_FILE};
	i2c_lock_write(EEPROM_ADR,buf,sizeof(buf));
}