/**
  ##############################################################################
	@file    modbus_regs_file.c
	@brief   Инициализацция и работа с регистровым файлом, для модбас
	@details Храним регистры в структуре modbus_regs_file, для каждого типа регистра есть своя "линия" (массив)
			Для использования, в начале нужно вызвать MBRF_init_start, "создать" (инициальзировать) регистры
			через MBRF_create_reg и вызвать MBRF_create_reg, ТОЛЬКО ПОСЛЕ этого разрешается ипользовать 
			функции чтения-записи регистров.
	@note   в текущей реализации в функциях чтения-записи нет проверки на то что инициализация файла была
			закончена, ответственность лежит на программисте. Стоит вызывать init_start и init_end в одном месте
  ##############################################################################
*/

#include "platform.h"
#include "modbus_regs_file.h"

typedef struct{
	uint16_t addr;               //register address
	uint16_t* data; 
}MB_register;

typedef struct{
	MB_register* line[4];
	TYPE_LINE_LEN len[4];
	TYPE_LINE_LEN len_max[4];
	bool create_reg_en_flag; //если флаг = 1 то разрешенно "создавать" регистры
}modbus_regs_file;

#define R_DATA(indx, type) (*(MB_R_file.line[(type)][(indx)].data))
#define R_DATA_POINTER(indx, type) (MB_R_file.line[(type)][(indx)].data)
#define R_DATA_BOOL(indx, type) (((R_DATA((indx_start+i),reg_type)) != 0u) ? 1u : 0u) //равно либо 0u либо 1u, страшно, но соответствует равилам misra
#define R_ADDR(indx,type) (MB_R_file.line[(type)][(indx)].addr)
#define R_LEN(type) (MB_R_file.len[(type)])

static void MBRF_swap(TYPE_LINE_LEN len_position_reg_1,TYPE_LINE_LEN len_position_reg_2, MB_register reg[]);
static MBR_status MBRF_selection_sort(MB_register reg[],TYPE_LINE_LEN len);
static MBR_status MBRF_binary_search(const MB_register reg[],TYPE_LINE_LEN len, uint16_t seach_addr, uint16_t* return_id);

static modbus_regs_file MB_R_file={.create_reg_en_flag=false};

/**
  @brief "начало" инициализации
  @brief инициализируем ключевую структуру MB_R_file и ставим create_reg_en_flag
 */
void MBRF_init_start(void)
{
	static MB_register holding_line[MAX_QUANTITY_HOLDING_REGS];
	static MB_register input_line[MAX_QUANTITY_INPUT_REGS];
	static MB_register input_discrete_line[MAX_QUANTITY_INPUT_DISCRETE_REGS];
	static MB_register colis_line[MAX_QUANTITY_COLIS_REGS];

	MB_R_file.line[MBR_holding]=holding_line;
	MB_R_file.line[MBR_input]=input_line;
	MB_R_file.line[MBR_input_discrete]=input_discrete_line;
	MB_R_file.line[MBR_coils]=colis_line;

	MB_R_file.len[MBR_holding]=0u;
	MB_R_file.len[MBR_input]=0u;
	MB_R_file.len[MBR_input_discrete]=0u;
	MB_R_file.len[MBR_coils]=0u;

	MB_R_file.len_max[MBR_holding]=MAX_QUANTITY_HOLDING_REGS;	
	MB_R_file.len_max[MBR_input]=MAX_QUANTITY_INPUT_REGS;
	MB_R_file.len_max[MBR_input_discrete]=MAX_QUANTITY_INPUT_DISCRETE_REGS;
	MB_R_file.len_max[MBR_coils]=MAX_QUANTITY_COLIS_REGS;

	MB_R_file.create_reg_en_flag=true; //разрешаем создавать регистры
	return;
}

/**
  @brief "Конец" инициализации
  @brief сортируем все "линии" регистров, выискивая дубликаты 
  @brief ставим create_reg_en_flag
  @retun MBR_ADDR_ERROR если встретили одинаковые адресса
  		 MBR_OK в нормальном случае
  @note текущая реализация отыскивает дубликаты только в рамках одной строки,
  @note например мы можем создать регистры с адрессом 0x23 типа coils и input одновременно
 */
MBR_status MBRF_init_end(void)
{
	MBR_status ret = MBR_OK;
	for(MBR_reg_type i=0;i<=MBR_coils;i++)
	{
		if(MBR_OK!=MBRF_selection_sort(MB_R_file.line[i], R_LEN(i)))
		{
			ret=MBR_ADDR_ERROR;
		}
	}
	MB_R_file.create_reg_en_flag=false; //запрещаем создавать регистры
	return ret;
}

/**
  @brief Инициализирует регистр
  @param addr modbus адресс регистра
  @param data указатель на данные которые в него положим
  @param тип регистра в который пишем 
  @retun MBR_CEATE_DISABLE если стоит блоировка
  		 MBR_LEN_ERROR если регистров этого типа в файле больше чем MAX_QUANTITY
  		 MBR_OK в нормальном случае
  		 MBR_ERROR в случае других ошибок
 */
MBR_status MBRF_create_reg(uint16_t addr, uint16_t* data, MBR_reg_type reg_type)
{
	MBR_status ret = MBR_ERROR;
	if(MB_R_file.create_reg_en_flag==true)
	{
		if(data!=0u)
		{
			if(R_LEN(reg_type)<MB_R_file.len_max[reg_type])
			{
				R_DATA_POINTER(R_LEN(reg_type),reg_type)=data;
				R_ADDR(R_LEN(reg_type),reg_type)=addr;
				R_LEN(reg_type)++;
				ret=MBR_OK;
			}else{
				ret=MBR_LEN_ERROR;
			}
		}
	}else{
		ret = MBR_CEATE_DISABLE;
	}
	return ret;
}

/**
  @brief Читаем данные из регистра (он должен быть инициализирован)
  @param read_addr modbus адресс регистра с которого читаем
  @param return_data по этому указателю запишем данные в случае успеха 
  @param тип регистра в который пишем 
  @retun MBR_OK в нормальном случае, если не нашли регистр то MBR_ERROR
 */
MBR_status MBRF_read_reg(uint16_t read_addr, uint16_t* return_data, MBR_reg_type reg_type)
{
	uint16_t indx;
	MBR_status ret = MBRF_binary_search(MB_R_file.line[reg_type],R_LEN(reg_type), read_addr, &indx);	
	if(ret==MBR_OK)
	{
		*return_data=R_DATA(indx,reg_type);
	}
	return ret;
}

/**
  @brief Читаем данные из регистров (они должен быть инициализированы и идти без "окон")
  @param read_addr_start modbus адресс первого регистра
  @param reg_len кол-во регистров
  @param return_data указатель на массив в который запишем результат
  @param тип регистра  
  @retun MBR_OK в нормальном случае, MBR_ERROR если не смогли найти 
  		 read_addr_start или read_addr_end или между ними есть "окно"
  @note в текущей реализации нету проверки на размер return_data
  @note например мы читаем 5 регистра начиная с 0x31:
  @note все 3 регистра должны существовать, MBR_LEN_ERROR если хотим прочитать 0 регистров, иначе return MBR_ERROR
 */
MBR_status MBRF_multiple_read_reg(uint16_t read_addr_start, uint16_t reg_len, uint8_t return_data[], MBR_reg_type reg_type)
{
	uint16_t indx_start; //indx_start_start не назван так, потому что в цикле потом с ним же работаем
	uint16_t indx_end;
	MBR_status ret=MBR_ERROR;
	if(reg_len>0u)
	{
		ret = MBRF_binary_search(MB_R_file.line[reg_type],R_LEN(reg_type), read_addr_start, &indx_start);	
		if(ret==MBR_OK)
		{
			ret=MBRF_binary_search(MB_R_file.line[reg_type],R_LEN(reg_type), read_addr_start+(reg_len-1u), &indx_end);
			if(MBR_OK==ret)
			{
				if((indx_end-indx_start)!=(reg_len-(uint16_t)1))
				{
					ret=MBR_ERROR;
				}else{
					for(TYPE_LINE_LEN i = 0; i<reg_len; i++)
					{
						return_data[i*2u]=(R_DATA((indx_start+i),reg_type))>>8u;
						return_data[(i*2u)+1u]=(R_DATA((indx_start+i),reg_type));
					}
					ret=MBR_OK;
				}
			}
		}
	}else{
		ret = MBR_LEN_ERROR;
	}
	return ret;
}

/**
  @brief Читаем данные из регистров в формате bool (они должен быть инициализированы и идти без "окон")
  @param read_addr_start modbus адресс первого регистра
  @param reg_len кол-во регистров 
  @param return_data указатель на массив в который запишем результат
  @param тип регистра  
  @retun MBR_OK в нормальном случае, MBR_ERROR если не смогли найти 
  		 read_addr_start или read_addr_end или между ними есть "окно"
  @note в текущей реализации нету проверки на размер return_data
  @note например мы читаем 5 регистра начиная с 0x31:
  @note все 3 регистра должны существовать, MBR_LEN_ERROR если хотим прочитать 0 регистров, иначе return MBR_ERROR
  @note текущая функция пишет данные в return_data[], в формате, нужном для 1 и 2 кодов функций modbus
  @note например если мы читает 2 регистра то фунция запишет 1байт, где 1вый бит это значение 1го регистра а 2ой второго
 */
MBR_status MBRF_multiple_read_reg_bool(uint16_t read_addr_start, uint16_t reg_len, uint8_t return_data[], MBR_reg_type reg_type)
{
	uint16_t indx_start; //indx_start_start не назван так, потому что в цикле потом с ним же работаем
	uint16_t indx_end;
	MBR_status ret=MBR_ERROR;
	if(reg_len>0u)
	{
		ret = MBRF_binary_search(MB_R_file.line[reg_type],R_LEN(reg_type), read_addr_start, &indx_start);	 
		if(ret==MBR_OK)
		{
			ret=MBRF_binary_search(MB_R_file.line[reg_type],R_LEN(reg_type), read_addr_start+(reg_len-1u), &indx_end);
			if(MBR_OK==ret) 
			{
				if((indx_end-indx_start)!=(reg_len-(uint16_t)1))
				{
					ret=MBR_ERROR;
				}else{ 
					for(uint8_t i =0;i<reg_len;i++)
					{
						return_data[i]=0u;
					}
					for(TYPE_LINE_LEN i = 0; i<reg_len; i++)
					{	
						return_data[i/8u]|=(R_DATA_BOOL((indx_start+i),reg_type))<<(i%8u); //читай Modbus_Application_Protocol_V1_1b3.pdf и будет понятно
					}
					ret=MBR_OK;
				}
			}
		}
	}else{
		ret = MBR_LEN_ERROR;
	}
	return ret;
}

/**
  @brief Заисываем данные в регистр (он должен быть инициализирован)
  @param addr modbus адресс регистра в который пишем
  @param data данные которые нужно записать
  @param тип регистра в который пишем 
  @retun MBR_OK в нормальном случае, если не нашли регистр то MBR_ERROR
 */
MBR_status MBRF_write_reg(uint16_t addr, uint16_t data, MBR_reg_type reg_type)
{
	uint16_t indx;
	MBR_status ret = MBRF_binary_search(MB_R_file.line[reg_type],R_LEN(reg_type), addr, &indx);	
	if(ret==MBR_OK)
	{
		R_DATA(indx,reg_type)=data;
	}
	return ret;
}

/**
  @brief Простая сортировка массива регистров + проверка на одинаковые адресса
  @param reg массив регистров
  @param len длинна массива 
  @retun MBR_OK в нормальном случае, если встретили одинаковые адресса то MBR_ADDR_ERROR
 */
static MBR_status MBRF_selection_sort(MB_register reg[],TYPE_LINE_LEN len)
{
	MBR_status ret = MBR_OK;
	for(TYPE_LINE_LEN i=0;i<len;i++)
	{
		TYPE_LINE_LEN min_idx=i;
		for(TYPE_LINE_LEN j=i+1u;j<len;j++)
		{
			if(reg[j].addr<reg[min_idx].addr)
			{
				min_idx=j;
			}else {
				if (reg[j].addr==reg[min_idx].addr)
				{
					ret = MBR_ADDR_ERROR;
				}
			}
		}
		if(min_idx != i)
		{
			MBRF_swap(min_idx,i,reg);
		}
	}
	return ret;
}

/**
  @brief бинарный поиск регистра с нужным адрессом в массиве
  @param reg массив регистров
  @param len длинна массива 
  @param addr ищем регистр с таким адрессом
  @param return_idx сюда положим индекс массива, если нашли
  @retun MBR_ERROR если не нашли и MBR_OK если нашли
 */
static MBR_status MBRF_binary_search(const MB_register reg[],TYPE_LINE_LEN len, uint16_t seach_addr, uint16_t* return_idx)
{
	MBR_status ret=MBR_ERROR;
	if(len>0u)
	{
		TYPE_LINE_LEN low=0;
		TYPE_LINE_LEN high=len-1u;
		while (low <= high) 
    {
    	TYPE_LINE_LEN mid = low + (high - low)/2u;
      if (reg[mid].addr == seach_addr)
      {
      	*return_idx=mid;
				ret=MBR_OK;
				break;
      }
      if ( (reg[mid].addr > seach_addr) && (mid!=0u) ) //больше и не дошли до конца
      {
      	high = mid - 1u;
      } else {
      	low = mid + 1u;
      }    
    }
	}
	return ret;
}

/**
  @brief  перестановка регистров местами
 */
static void MBRF_swap(TYPE_LINE_LEN len_position_reg_1,TYPE_LINE_LEN len_position_reg_2, MB_register reg[])
{	
	MB_register bud_reg; //используем этот тип данных для всех типов регистров
	bud_reg.data = reg[len_position_reg_1].data; //дата и аддрес 1го в буфер
	bud_reg.addr = reg[len_position_reg_1].addr;
	reg[len_position_reg_1].data = reg[len_position_reg_2].data; //дата и аддрес 1го = дата и аддрес 2го
	reg[len_position_reg_1].addr = reg[len_position_reg_2].addr;
	reg[len_position_reg_2].data = bud_reg.data; //дата и аддрес 1го = буфер
	reg[len_position_reg_2].addr = bud_reg.addr; 
}


#ifdef UTEST_BUILD
/**
  @brief Очистка файла регистров. нужен только для тестов
 */
void MBRF_clear_file(void)
{
	for(TYPE_LINE_LEN i=0;(MBR_reg_type)i<=MBR_coils;i++)
	{
		for(TYPE_LINE_LEN j=0;j<R_LEN(i);j++)
		{
        	//printf("\n\n[d]=%04x [a]=%04x\n",R_DATA(j,i),R_ADDR(j,i));
			R_DATA(j,i)=0u;
			R_ADDR(j,i)=0u;
		}
		R_LEN(i)=0u;
	}
}

TYPE_LINE_LEN MBRF_MAX_QUANTITY(MBR_reg_type reg_type)
{
	return MB_R_file.len_max[reg_type];
}
#endif