/**
  ******************************************************************************
  * @file    i2c.c
  * @brief   измерение температуры
  * @note    написано на скорость, нужно processing_the_measurement_state доисать 
  * @note    и добавит поддержу других сенсоров и режимы работы с подогревом
  ******************************************************************************
  */

#include "temp.h"
#include "i2c.h"
#include "crc.h"
#include "modbus_regs_file.h"
#include "sys_tim.h"
#include "state_machine.h"
#include "modbus_config.h"

#define SHT4X_ADR 0x44

#define DELAY_MEASUREMENT 976u // примерно 1000 с поправкой на DELAY_AFTER_PROCESSING и задержки в main
#define DELAY_AFTER_PROCESSING 10u

//Command
#define C_MEASURE_HIGH 0xFD
#define C_MEASURE_MEDIUM 0xF6
#define C_MEASURE_LOWEST 0xE0
#define C_SERIAL_NUMBER 0x89
#define C_SOFT_RESET 0x94
#define HEATER_200MW_1S 0x39
#define HEATER_200MW_01S 0x32
#define HEATER_110MW_1S 0x2F
#define HEATER_110MW_01S 0x24
#define HEATER_20MW_1S 0x1E
#define HEATER_20MW_01S 0x15

static void processing_the_measurement_state(state_machine_dev* dev);
static void read_the_measurement_state(state_machine_dev* dev);
static void init_the_measurement_state(state_machine_dev* dev);
static void ic_reset(void);

static uint16_t temperature = MBDDATA_TEMPERATURE;
static uint16_t temperature_compensation = MBDDATA_TEMPERATURE_COMPENSATION;
static uint16_t humidity = MBDDATA_HUMIDITY;

static uint8_t i2c_buf[8] = {0};

static state_machine_dev st_dev = {plug_state, plug_state, 0u};

/**
   @brief init
 */
void temp_sensor_init(void)
{
	ic_reset();
	set_state_after_delay(init_the_measurement_state, DELAY_AFTER_PROCESSING, &st_dev);
}

/**
   @brief работаем
 */
void temperature_sensor_periodic_work(void)
{
	execute_current_state(&st_dev);
	return;
}

/**
   @brief обработка данных
   @note тестовый вариант
   @note нужно добавить проверку CRC и все вычисления обдумать
 */
static void processing_the_measurement_state(state_machine_dev* dev)
{
	//мы не проверяем CRC, нужно исправить

	uint16_t te = (i2c_buf[0]<<8)+i2c_buf[1];
	float f_data = ((-45 + (175 * (te / 65536.0)))/0.1); //переисать в норманый вид
	uint16_t data = (uint16_t)(f_data-((float)(temperature_compensation)*0.1));
	temperature = data;
    int te1 = (i2c_buf[3]<<8)+i2c_buf[4];
	int data1 = (int)((-6 + 125 * te1/65535.0)/0.1); //переисать в норманый вид
	humidity = data1;
    set_state_after_delay(init_the_measurement_state, DELAY_MEASUREMENT, &st_dev);

    return;
}

/**
   @brief читаем измерение
 */
static void read_the_measurement_state(state_machine_dev* dev)
{
	if(I2C_NOT_BUSY==i2c_read(SHT4X_ADR,&i2c_buf,6u))
	{
		set_state_after_delay(processing_the_measurement_state, 10u, &st_dev);
	}
	return;
}

/**
   @brief посылаем запрос на измерение
 */
static void init_the_measurement_state(state_machine_dev* dev)
{
	if(I2C_NOT_BUSY==i2c_write(SHT4X_ADR, i2c_buf, 1))
	{
		set_state_after_delay(read_the_measurement_state, 10u, &st_dev);
	}
	return;
}

/**
   @brief перезагружаем микросхему
 */
static void ic_reset(void)
{
	i2c_write(SHT4X_ADR, i2c_buf, 1);
	return;
}

/**
   @brief передаём указатели в модбас регистры
 */
uint8_t MBRF_temperature_sensor_init(void)
{
	uint8_t MB_init_eror=0u;
	MB_init_eror+=MBRF_create_reg(MBADDR_TEMPERATURE, &temperature, MBTYPE_TEMPERATURE); //0x0
	MB_init_eror+=MBRF_create_reg(MBADDR_TEMPERATURE_COMPENSATION, &temperature_compensation, MBTYPE_TEMPERATURE_COMPENSATION); //0x00F5
	MB_init_eror+=MBRF_create_reg(MBADDR_HUMIDITY, &humidity, MBTYPE_HUMIDITY); //0x0
	return MB_init_eror;
}
