/**
  ##############################################################################
  @file     main.c
  @author   Борисков Кирилл
  @brief    Архитектура прошивки крайне проста, 
  @brief    работаем с i2c и uart по DMA, в цикле обрабатываем данные
  ##############################################################################
*/

#include "dout.h"
#include "uart_debug.h"
#include "xprintf.h"
#include "led.h"
#include "rcc.h"
#include "sys_tim.h"
#include "uart_modbus.h"
#include "modbus_regs_file.h"
#include "modbus_config.h"
#include "i2c.h"

#include "temp.h"

static void modbus_struc_init(void);
void check_reset_flag(void);

static uint16_t reset_flag=MBDDATA_RESET;

int main(void)
{
  sysclk_LSI_setup(); //для работы собаки
  watchdog_init(); //на 0.5 сек
  sysclk_64mhz_from_hsi16_setup();
  SysTick_Init(); 
  
  #ifdef DEBUG
  uart_debug_init();
  xprintf("\n\n\\\\\\\\\\\\\\\\\\\\\\\\\\\\\n   START\n//////////////\n");
  #endif

  i2c_init();
  temp_sensor_init();
  dout_init();
  led_config();

  modbus_struc_init(); 
  eeprom_file_init();
  uart_modbus_init();

  watchdog_refresh();

  #ifdef DEBUG
  xprintf("//////////////\n   INIT SUCCESSFULLY\n uptime: %dms (%dsec)\n//////////////\n",time_get_ms(),time_get_sec());
  #endif

  while(1)
  {
    modbus_periodic_work();
    check_reset_flag();

    temperature_sensor_periodic_work();
    led_periodic_work();
    dout_periodic_work();
    eeprom_periodic_work();

    delay(1);
    watchdog_refresh();
  }

  return 0;
}

/**
   @brief инициализируем структуру modbus регистров перезагрузка МК в случае проблеммы
 */
static void modbus_struc_init(void)
{
  uint8_t MB_init_eror = 0u;
  MBRF_init_start();

  uint8_t buf_model[] = {MBDDATA_MODEL};
  //Модель устройства (строка asci "TempControl") адресса 200-205
  for(uint8_t i =0; i<sizeof(buf_model);++i)
  {
    MB_init_eror+=MBRF_create_reg(MBADDR_MODEL+i, &buf_model[i], MBTYPE_MODEL);
  }
  
  MB_init_eror+=MBRF_create_reg(MBADDR_RESET, &reset_flag, MBTYPE_RESET);
  MB_init_eror+=MBRF_led_init();
  MB_init_eror+=MBRF_dout_init();
  MB_init_eror+=MBRF_modbus_init();
  MB_init_eror+=MBRF_temperature_sensor_init();
  MB_init_eror+=MBRF_time_init();

  #ifdef DEBUG //заглушка для дебага
  static uint16_t humidity_plug = 0x7FFF;
  MB_init_eror+=MBRF_create_reg(0x2, &humidity_plug, MBR_input); //0x2 заглушка reg_illumination
  #endif

  if((MB_init_eror!=MBR_OK)||(MBR_OK!=MBRF_init_end()))
  {
    #ifdef DEBUG
    xprintf("MBRF_init_end eror\n");
    #endif
    RESET_MK(); //это очень спорно нужно обдумать //по идее это тольо на столе при дебаге может случиться
  }else{
    #ifdef DEBUG
    xprintf("modbus file init\n");
    #endif
  }

  xprintf("\n");
  return;
}


/**
   @brief проверяем reset_flag и перезагружаем МК 
 */
void check_reset_flag(void)
{
    if(reset_flag!=0x0) 
    {
      //watchdog_refresh();  //по тестам мы успеваем ответить даже на 1200бод
      while(MBRF_check_transfer_not_completed()) //ждём пока отет о успешной записи в reset_flag доедёт
      {
        NOP();
      }
      RESET_MK(); //перезагружаем МК
    }
}


