/**
  ******************************************************************************
  * @file    i2c.c
  * @brief   i2c functions
  ******************************************************************************
  */

#include "platform.h"
#include "i2c.h"
#include "bool.h"

#include "xprintf.h"//delete
#include "sys_tim.h"//delete
#include "led.h" //delete

#define CHECK_NOT_BUSY() ((I2C2->ISR & I2C_ISR_BUSY)!=I2C_ISR_BUSY)

static void i2c_DMA_init_read(const uint8_t* data, uint8_t length);
static void i2c_DMA_init_write(const uint8_t* data, uint8_t length);
static void i2c_DMA_init(void);

static i2c_status last_transmit_nack =I2C_ERROR;

/**
   @brief настройка i2c  
   @note текущая реализация расчитана на 64мгц тактирование
 */
#define DMAREQ_ID_I2C2_RX 12u
#define DMAREQ_ID_I2C2_TX 13u
i2c_status i2c_init(void)
{
    RCC->APBENR1 |= RCC_APBENR1_I2C2EN; // Включаем тактирование I2C2
    RCC->IOPENR  |= RCC_IOPENR_GPIOAEN; // Включаем тактирование GPIO порта A

    // Настройка PA11 и PA12 
    GPIOA->MODER &= ~(GPIO_MODER_MODE11_0 | GPIO_MODER_MODE12_0); // Сбросить биты
    GPIOA->MODER |= (GPIO_MODER_MODE11_1 | GPIO_MODER_MODE12_1);  // Установить режим альтернативной функции

    GPIOA->OTYPER  |= GPIO_OTYPER_OT12 | GPIO_OTYPER_OT11;
    GPIOA->PUPDR   &= ~(GPIO_PUPDR_PUPD11_1 | GPIO_PUPDR_PUPD12_1);
    GPIOA->PUPDR   |= (GPIO_PUPDR_PUPD11_0 | GPIO_PUPDR_PUPD12_0);

    GPIOA->AFR[1] |= (GPIO_AFRH_AFSEL12_1 | GPIO_AFRH_AFSEL12_2 | GPIO_AFRH_AFSEL11_1 | GPIO_AFRH_AFSEL11_2); // Установить альтернативную функцию AF1 

    GPIOA->OSPEEDR |=(GPIO_OSPEEDR_OSPEED12 | GPIO_OSPEEDR_OSPEED11); // Настройка скорости GPIO

    I2C2->CR1 &= ~I2C_CR1_PE; // Отключаем I2C на время настройки
    I2C2->CR1 &= ~I2C_CR1_ANFOFF; // Включаем аналоговый фильтр
    I2C2->CR1 &= ~I2C_CR1_DNF; // Отключаем цифровой фильтр
    I2C2->CR1 |= I2C_CR1_TXDMAEN; //DMA transmission requests enable 
    I2C2->CR1 |= I2C_CR1_RXDMAEN;
    I2C2->CR1 |= I2C_CR1_TCIE; //Transfer complete interrupt enable 
    I2C2->CR1 |= I2C_CR1_NACKIE; //NACK received interrupt enable
    I2C2->CR1 &= ~I2C_CR1_NOSTRETCH; // Отключение растяжения тактового сигнала

    // // Настройка таймингов, 100 кГц при 64 МГц тактировании 
    // //значения из 25.4.11 I2C_TIMINGR register configuration examples меняем только PRESC
    // I2C2->TIMINGR |= (0x13 << I2C_TIMINGR_SCLL_Pos) | 
    //               (0xf << I2C_TIMINGR_SCLH_Pos) | 
    //               (0x2 << I2C_TIMINGR_SDADEL_Pos) | 
    //               (0x4 << I2C_TIMINGR_SCLDEL_Pos) | 
    //               (0x1D << I2C_TIMINGR_PRESC_Pos); 

    // Настройка таймингов, 35 кГц при 64 МГц тактировании
    I2C2->TIMINGR |= (0x55 << I2C_TIMINGR_SCLL_Pos) | 
                  (0x51 << I2C_TIMINGR_SCLH_Pos) | 
                  (0x2 << I2C_TIMINGR_SDADEL_Pos) | 
                  (0x4 << I2C_TIMINGR_SCLDEL_Pos) | 
                  (0x9 << I2C_TIMINGR_PRESC_Pos); 

    I2C2->CR1 |= I2C_CR1_PE; // Включаем I2C
    NVIC_EnableIRQ(I2C2_IRQn);
    NVIC_SetPriority(I2C2_IRQn, I2C_IRQ_PIOITY_LEVEL);
    i2c_DMA_init();

    return I2C_OK;
}

/**
   @brief прерывания i2c 
 */
extern void I2C2_IRQHandler(void)
{   
    if(I2C2->ISR & I2C_ISR_TC )
    {
        I2C2->CR2 |= I2C_CR2_STOP;
        last_transmit_nack=I2C_ACK;
    }
    
    if(I2C2->ISR & I2C_ISR_NACKF )
    {
        I2C2->CR2 |= I2C_CR2_STOP;
        I2C2->ICR |= I2C_ICR_NACKCF;

        DMA1_Channel1->CCR &=~ DMA_CCR_EN;
        last_transmit_nack=I2C_NACK;
    }

    return;
}

/**
   @brief прерываниe dma 
 */
extern void DMA1_Channel1_IRQHandler(void)
{
    if(DMA1->ISR & DMA_ISR_TCIF1 )
    {
        DMA1->IFCR |= DMA_IFCR_CTCIF1;
        DMA1_Channel1->CCR &=~ DMA_CCR_EN;
    }
    return;
}

/**
  @brief запускаем dma на чтение из массива
  @param data указатель на массив откуда пишем в i2c
  @param length длина data
 */
static void i2c_DMA_init_read(const uint8_t* data, uint8_t length)
{
    DMA1_Channel1->CPAR = (&(I2C2->RXDR)); //указатель на передатчик
    DMA1_Channel1->CMAR = data; //адресс на начало буфера
    DMA1_Channel1->CNDTR = length; //счётчик
    DMAMUX1_Channel0->CCR = (DMAREQ_ID_I2C2_RX&DMAMUX_CxCR_DMAREQ_ID);
    DMA1_Channel1->CCR &=~ DMA_CCR_DIR; //Data transfer direction
    DMA1_Channel1->CCR |= DMA_CCR_EN;
    return;
}

/**
  @brief запускаем dma на запись в массив
  @param data указатель на массив куда пишем из i2c
  @param length длина data
 */
static void i2c_DMA_init_write(const uint8_t* data, uint8_t length)
{
    DMA1_Channel1->CPAR = (&(I2C2->TXDR)); //указатель на передатчик
    DMA1_Channel1->CMAR = data; //адресс на начало буфера
    DMA1_Channel1->CNDTR = length;         /*Set number of bytes to transmit*/
    DMAMUX1_Channel0->CCR = (DMAREQ_ID_I2C2_TX&DMAMUX_CxCR_DMAREQ_ID);
    DMA1_Channel1->CCR |= DMA_CCR_DIR; //Data transfer direction
    DMA1_Channel1->CCR |= DMA_CCR_EN; 
    return;
}

/**
  @brief начальная настройка dma 
 */
static void i2c_DMA_init(void)
{
    RCC->AHBENR  |= RCC_AHBENR_DMA1EN; //включаем тактирование 
    DMA1_Channel1->CCR |= (DMA_CCR_PL  //priority level hery high
                    | DMA_CCR_MINC //memory increment mode enable
                    | DMA_CCR_TCIE //transfer complete interrupt enable
                    //| DMA_CCR_TEIE //Transfer error interrupt enable 
                    );

    NVIC_SetPriority(DMA1_Channel1_IRQn, I2C_IRQ_PIOITY_LEVEL);
    NVIC_EnableIRQ(DMA1_Channel1_IRQn);
    DMA1_Channel1->CCR &=~ DMA_CCR_EN;
}

/**
   @brief чтение регистров по DMA  
   @param addr адресс i2c slave 
   @param data указатель на массив куда пишем из i2c
   @param length длина data
   @return I2C_ERROR - непонятно что случилось
   @return I2C_NOT_BUSY - шина не занята, чтение по дма запущенно
   @return I2C_BUSY - шина не занята, ждите
 */
i2c_status i2c_read(uint8_t addr, uint8_t* data, uint8_t length)
{
    i2c_status ret = I2C_ERROR;
    if (length < 1u)
    {
        return I2C_LENGTH_ERROR; //не заданна длина data[]
    }

    if(CHECK_NOT_BUSY())
    {
        ret = I2C_NOT_BUSY;
        I2C2->CR2 = ((addr << 1U) & I2C_CR2_SADD) | I2C_CR2_RD_WRN | (length << I2C_CR2_NBYTES_Pos);
        i2c_DMA_init_read(data, length);
        I2C2->CR2 |= I2C_CR2_START;
    }else{
        ret = I2C_BUSY;
    }

    return ret;
}

/**
   @brief запись регистров по DMA  
   @param addr адресс i2c slave 
   @param data указатель на массив куда пишем из i2c
   @param length длина data
   @return I2C_ERROR - непонятно что случилось
   @return I2C_NOT_BUSY - шина не занята, запись по дма запущенна
   @return I2C_BUSY - шина не занята, ждите
 */
i2c_status i2c_write(uint8_t addr, uint8_t* data, uint8_t length)
{
    i2c_status ret = I2C_ERROR;
    if (length < 1u)
    {
        return I2C_LENGTH_ERROR; //не заданна длина data[]
    }

    if(CHECK_NOT_BUSY())
    {
        ret = I2C_NOT_BUSY;
        I2C2->CR2 &=~ (I2C_CR2_SADD) | (I2C_CR2_NBYTES) | I2C_CR2_RD_WRN;
        I2C2->CR2 = ((addr << 1U) & I2C_CR2_SADD) | (length << I2C_CR2_NBYTES_Pos);
        i2c_DMA_init_write(data, length);
        I2C2->CR2 |= I2C_CR2_START;
    }else{
        ret = I2C_BUSY;
    }

    return ret;
}

/**
   @brief проверка, занята шина или нет
   @return I2C_NOT_BUSY - шина не занята, запись по дма запущенна
   @return I2C_BUSY - шина не занята, ждите
 */
i2c_status i2c_check_busy(void)
{
    i2c_status ret = I2C_ERROR;
    if(CHECK_NOT_BUSY())
    {
        ret=I2C_NOT_BUSY;
    }else{
        ret=I2C_BUSY;
    }
    return ret;
}

/**
   @brief что выдало последние действия с i2c
   @return I2C_ACK 
   @return I2C_NACK 
 */
i2c_status i2c_check_last_transmit_ACK(void)
{
    return last_transmit_nack;
}

/**
   @brief блокирущее чтение регистров 
   @param addr адресс i2c slave 
   @param data указатель на массив куда пишем из i2c
   @param length длина data
   @return I2C_ERROR - не смогла я
   @return I2C_OK - прочитали успешно
   @note т.к. мы используем блокирущие функции только 
   @note во время инициалиации, принятно решение использовать
   @note обёртки на DMA функции
   @note 
   @note функция потенциально может заcтрять в бесконечном цикле
 */
i2c_status i2c_lock_read(uint8_t addr, uint8_t* data, uint8_t length)
{
    i2c_status ret=I2C_ERROR;
    while(i2c_check_busy()==I2C_BUSY)
    {
        NOP();
    }
    i2c_read( addr, data, length);
    while(i2c_check_busy()==I2C_NOT_BUSY) //ждём запуска дма
    {
        NOP();
    }
    while(i2c_check_busy()==I2C_BUSY)
    {
        NOP();
    }
    if(I2C_ACK==i2c_check_last_transmit_ACK())
    {
        ret=I2C_OK;
    }
    return ret;
}

/**
   @brief блокирущая запись регистров 
   @param addr адресс i2c slave 
   @param data указатель на массив откуда пишем в i2c
   @param length длина data
   @return I2C_ERROR - не смогла я
   @return I2C_OK - прочитали успешно
   @note т.к. мы используем блокирущие функции только 
   @note во время инициалиации, принятно решение использовать
   @note обёртки на DMA функции
   @note 
   @note функция потенциально может заcтрять в бесконечном цикле
 */
i2c_status i2c_lock_write(uint8_t addr, uint8_t* data,uint8_t length)
{
    i2c_status ret=I2C_ERROR;
    while(i2c_check_busy()==I2C_BUSY)
    {
        NOP();
    }
    i2c_write( addr, data, length);
    while(i2c_check_busy()==I2C_NOT_BUSY)
    {
        NOP();
    }
    while(i2c_check_busy()==I2C_BUSY)
    {
        NOP();
    }
    if(I2C_ACK==i2c_check_last_transmit_ACK())
    {
        ret=I2C_OK;
    }
    return ret;
}

#ifdef FFF //не используем этот код
i2c_status i2c_ACK(uint8_t addr)
{
    bool ret =false; //0-error 1-sess

    I2C2 -> CR2 |= I2C_CR2_NBYTES; //The number of bytes to be transmitted/received is programmed there.
    I2C2 -> CR2 |= I2C_CR2_RELOAD; // software end mode: TC flag is set when NBYTES data are transferred, stretching SCL low.

    I2C2 -> CR2 &= ~I2C_CR2_SADD;
    I2C2 -> CR2 |= ((addr<<1U)&I2C_CR2_SADD); //Slave address

    //I2C2->ISR &= ~I2C_ISR_NACKF;

    I2C2 -> CR2 |= I2C_CR2_START; //Start generation

    delay(1);
    //xprintf("Adr = 0x%02x ISR = 0x%08x\n",data,I2C2->ISR);
    if( !((I2C2->ISR&I2C_ISR_NACKF)==I2C_ISR_NACKF) )
    {
        ret = true; //0-error 1-sess
    }

    I2C2->CR2 |= I2C_CR2_STOP;

    //xprintf("0x%08x\n",((data<<1U)&I2C_CR2_SADD));
    return ret;
}
#endif